"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
downloading functions
"""

import logging
from pathlib import Path
import urllib.request
from osm_flex.config import DICT_GEOFABRIK, GEOFABRIK_URL, PLANET_URL, OSM_DATA_DIR

LOGGER = logging.getLogger(__name__)

# =============================================================================
#  DOWNLOAD METHODS
# =============================================================================

def _create_gf_download_url(iso3, file_format):
    """
    create string with download-url from geofabrik

    Parameters
    ----------
    iso3 : str
        ISO3 code of country to download
    file_format : str
        Format in which file should be downloaded; ESRI Shapefiles ('shp')
        or osm-Protocolbuffer Binary Format ('pbf')

    Returns
    -------
    str : Geofabrik download-string for the requested country.

    See also
    --------
    DICT_GEOFABRIK for exceptions / special regions.
    """
    try:
        if file_format == 'shp':
            return f'{GEOFABRIK_URL}{DICT_GEOFABRIK[iso3][0]}/{DICT_GEOFABRIK[iso3][1]}-latest-free.shp.zip'
        if file_format == 'pbf':
            return f'{GEOFABRIK_URL}{DICT_GEOFABRIK[iso3][0]}/{DICT_GEOFABRIK[iso3][1]}-latest.osm.pbf'
    except KeyError:
        if iso3=='RUS':
            raise KeyError("""Russia comes in two files. Please specify either
                         'RUS-A for the Asian or RUS-E for the European part.""")
        raise KeyError("""The provided iso3 seems not to be available on
                           Geofabrik.de. You can clip it from the planet
                           file or an adequate regional file, instead. See
                           the methods in the clip module for this.""")

    return LOGGER.error('invalid file format. Please choose one of [shp, pbf]')

# TODO: decide whether to issue warnings for multi-country files
def get_country_geofabrik(iso3, file_format='pbf', save_path=OSM_DATA_DIR,
                          overwrite=False):
    """
    Download country files with all OSM map info from the provider
    Geofabrik.de.

    Parameters
    ----------
    iso3 : str
        ISO3 code of country to download
        Exceptions: Russia is divided into European and Asian part
        ('RUS-E', 'RUS-A'), Canary Islands are 'IC'.
    file_format : str
        Format in which file should be downloaded; options are
        ESRI Shapefiles (shp), which can easily be loaded into gdfs,
        or osm-Protocolbuffer Binary Format (pbf), which is smaller in
        size, but has a more complicated query syntax to load (functions
        are provided in the OSMFileQuery class).
    save_path : str or pathlib.Path
        Folder in which to save the file

    Returns
    -------
    None
        File is downloaded and stored under save_path + the Geofabrik filename

    See also
    --------
    DICT_GEOFABRIK for exceptions / special regions.
    """

    download_url = _create_gf_download_url(iso3, file_format)
    local_filepath = Path(save_path , download_url.split('/')[-1])
    if not Path(local_filepath).is_file() or overwrite:
        LOGGER.info(f'Downloading file as {local_filepath}')
        urllib.request.urlretrieve(download_url, local_filepath)
    else:
        LOGGER.info(f'file already exists as {local_filepath}')

# TODO: allow for several spelling options like "Central America", "Australia", ...
def get_region_geofabrik(region, save_path=OSM_DATA_DIR, overwrite=False):
    """
    Download regions files with all OSM map info from the provider
    Geofabrik.de
    
    Parameters
    ----------
    region: str
        one of Africa, Antarctica, Asia, Australia-and-Oceania,
        Central-America, Europe, North-America, South-America
    save_path : str or pathlib.Path
        Folder in which to save the file
    """

    download_url =  f'{GEOFABRIK_URL}{region.lower()}-latest.osm.pbf'
    local_filepath = Path(save_path , download_url.split('/')[-1])
    if not Path(local_filepath).is_file() or overwrite:
        LOGGER.info(f'Downloading file as {local_filepath}')
        urllib.request.urlretrieve(download_url, local_filepath)
    else:
        LOGGER.info(f'file already exists as {local_filepath}')


def get_planet_file(save_path=Path(OSM_DATA_DIR,'planet-latest.osm.pbf'),
                    overwrite=False):
    """
    Download the entire planet file from the OSM server (ca. 60 GB).

    Parameters
    ----------
    save_path : str or pathlib.Path
    """

    if not Path(save_path).is_file() or overwrite:
        LOGGER.info(f'Downloading file as {save_path}')
        urllib.request.urlretrieve(PLANET_URL, save_path)
    else:
        LOGGER.info(f'file already exists as {save_path}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
init file
"""

from pathlib import Path
from .config import OSM_DIR, OSM_DATA_DIR, POLY_DIR, EXTRACT_DIR


def setup_osm_dirs():
    """
    This function is called when osm-flex is imported.
    It creates an osm directory by default in the home directory,
    with three sub-folders osm_bpf, poly and extract, for raw data dumps,
    .poly files, and extraction results, respectively.
    Other locations can be configured in the osm-flex.config file.
    
    Parameters
    ----------
    """
    OSM_DIR.mkdir(parents=True, exist_ok=True)
    for dirpath in [OSM_DATA_DIR, POLY_DIR, EXTRACT_DIR]:
        Path(dirpath).mkdir(parents=True, exist_ok=True)

setup_osm_dirs()



"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
test download functions
"""

import unittest
import tempfile
import os
import shapely
from pathlib import Path
from osm_flex.download import (_create_gf_download_url, 
                               get_country_geofabrik, 
                               get_region_geofabrik, 
                               get_planet_file)

class TestDownload(unittest.TestCase):

    def setUp(self):
        pass

    def test_create_gf_download_url(self):
        
        # TODO: assert raises KeyError
        #iso3 = 'ATG'
        #file_format = 'pbf'
        #_create_gf_download_url(iso3, file_format)
        
        # TODO: assert raises KeyError
        #iso3 = 'DEU'
        #file_format = 'pdf'
        #_create_gf_download_url(iso3, file_format)
        
        
        iso3 = 'DEU'
        file_format = 'pbf'
        url = _create_gf_download_url(iso3, file_format)
        self.assertEqual(url, 'https://download.geofabrik.de/europe/germany-latest.osm.pbf')
        
        iso3 = 'DEU'
        file_format = 'shp'
        url = _create_gf_download_url(iso3, file_format)
        self.assertEqual(url, 'https://download.geofabrik.de/europe/germany-latest-free.shp.zip')
        
        iso3 = 'RUS-A'
        file_format = 'pbf'
        url = _create_gf_download_url(iso3, file_format)
        self.assertEqual(url, 'https://download.geofabrik.de/asia/russia-latest.osm.pbf')
        
        iso3 = 'RUS-E'
        file_format = 'pbf'
        url = _create_gf_download_url(iso3, file_format)
        self.assertEqual(url, 'https://download.geofabrik.de/europe/russia-latest.osm.pbf')
    
        
    def test_get_country_geofabrik(self):
        
        iso3 = 'IC'
        with tempfile.TemporaryDirectory() as tmpdir:
            filename = os.path.join(tmpdir, "canary-islands-latest.osm.pbf")
            get_country_geofabrik(iso3, save_path=tmpdir)
            self.assertTrue(os.path.exists(filename))

    def test_get_region_geofabrik(self):

        region = 'central-america'
        with tempfile.TemporaryDirectory() as tmpdir:
            filename = os.path.join(tmpdir, "central-america-latest.osm.pbf")
            get_region_geofabrik(region, save_path=tmpdir)
            self.assertTrue(os.path.exists(filename))

    def test_get_planet_file(self):
        pass

if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestDownload)
    unittest.TextTestRunner(verbosity=2).run(TESTS)




"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
test clip functions
"""

import unittest
import tempfile
import os
import shapely
from pathlib import Path
from osm_flex.clip import (get_admin1_shapes, get_country_shape, 
                           _simplify_shapelist, _shapely2poly, _build_osmosis_cmd,
                           _build_osmconvert_cmd)
from osm_flex.config import OSMCONVERT_PATH


class TestClip(unittest.TestCase):
    def test_get_admin1_shapes(self):
        # Test for valid country code
        result = get_admin1_shapes("CAN")
        self.assertIsInstance(result, dict)
        self.assertGreater(len(result), 0)
        self.assertIn("Alberta", result.keys())

        # Test for invalid country code
        with self.assertRaises(LookupError):
            get_admin1_shapes("XYZ")

        # Test for invalid input type
        with self.assertRaises(TypeError):
            get_admin1_shapes(123)

    def test_get_country_shape(self):
        # Test for valid country code
        result = get_country_shape("CAN")
        self.assertIsInstance(result, shapely.geometry.base.BaseGeometry)

        # Test for invalid country code
        with self.assertRaises(LookupError):
            get_country_shape("XYZ")

        # Test for invalid input type
        with self.assertRaises(TypeError):
            get_country_shape(123)

    def test__simplify_shapelist(self):
        # Test for valid input
        geom_list = [shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])]
        result = _simplify_shapelist(geom_list)
        self.assertEqual(result[0].wkt, 'POLYGON ((0 0, 1 0, 1 1, 0 1, 0 0))')

        # Test too small valid input
        geom_list = [shapely.geometry.Polygon([(0, 0), (0.01, 0), (0.01, 0.01), (0, 0.01)])]
        result = _simplify_shapelist(geom_list)
        self.assertFalse(result) #check if result is empty list

        # Test for invalid input type
        with self.assertRaises(TypeError):
            _simplify_shapelist("invalid_input")

    def test__shapely2poly(self):
        # Test for valid input
        geom_list = [shapely.geometry.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])]
        with tempfile.TemporaryDirectory() as tmpdir:
            filename = os.path.join(tmpdir, "test_poly")
            _shapely2poly(geom_list, filename)
            with open(filename+".poly", "r") as f:
                content = f.read()
                self.assertIn("Polygons", content)
                self.assertIn("0", content)
                self.assertIn("0.0     0.0", content)
                self.assertIn("1.0     0.0", content)
                self.assertIn("1.0     1.0", content)
                self.assertIn("0.0     1.0", content)
                self.assertIn("END", content)
        # Test for invalid input type
        with tempfile.TemporaryDirectory() as tmpdir:
            filename = os.path.join(tmpdir, "test_poly")
            with self.assertRaises(AttributeError):
                _shapely2poly("invalid_input", filename)

    def test__build_osmosis_cmd(self):
        # Test for bbox input
        shape = [0, 0, 1, 1]
        osmpbf_clip_from = "/path/to/planet.osm.pbf"
        osmpbf_output = "/path/to/extract.osm.pbf"
        result = _build_osmosis_cmd(shape, osmpbf_clip_from, osmpbf_output)
        expected_result = [
            'osmosis',
             '--read-pbf',
             'file=/path/to/planet.osm.pbf',
             '--bounding-box',
             'top=1',
             'left=0',
             'bottom=0',
             'right=1',
             '--write-pbf',
             'file=/path/to/extract.osm.pbf']
        self.assertEqual(result, expected_result)
        
        # Test for .poly file input
        shape = "/path/to/shape.poly"
        osmpbf_clip_from = "/path/to/planet.osm.pbf"
        osmpbf_output = "/path/to/extract.osm.pbf"
        result = _build_osmosis_cmd(shape, osmpbf_clip_from, osmpbf_output)
        expected_result = ['osmosis',
         '--read-pbf',
         'file=/path/to/planet.osm.pbf',
         '--bounding-polygon',
         'file=/path/to/shape.poly',
         '--write-pbf',
         'file=/path/to/extract.osm.pbf']
        self.assertEqual(result, expected_result)
        
    def test__build_osmconvert_cmd(self):
        # Test for valid input
        shape = [0, 0, 1, 1]
        osmpbf_clip_from = "/path/to/planet.osm.pbf"
        osmpbf_output = "/path/to/extract.osm.pbf"
        result = _build_osmconvert_cmd(shape, osmpbf_clip_from, osmpbf_output)
        expected_result = [str(OSMCONVERT_PATH),
         '/path/to/planet.osm.pbf',
         '-b=0,0,1,1',
         '-o=/path/to/extract.osm.pbf']
        self.assertEqual(result, expected_result)
        
        # Test for .poly file input
        shape = Path("/path/to/shape.poly")
        osmpbf_clip_from = "/path/to/planet.osm.pbf"
        osmpbf_output = "/path/to/extract.osm.pbf"
        result = _build_osmconvert_cmd(shape, osmpbf_clip_from, osmpbf_output)
        expected_result = [str(OSMCONVERT_PATH),
         '/path/to/planet.osm.pbf',
         '-B=/path/to/shape.poly',
         '-o=/path/to/extract.osm.pbf']
        self.assertEqual(result, expected_result)
        


if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestClip)
    unittest.TextTestRunner(verbosity=2).run(TESTS)



"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
test simplification functions
"""

import unittest
import geopandas as gpd
import shapely as sh
from osm_flex.simplify import (remove_small_polygons, remove_contained_points,
                               remove_contained_polys, remove_exact_duplicates)
from pandas.testing import assert_frame_equal
from pathlib import Path

PATH_TEST_DATA = Path(__file__).parent / 'data'
OSM_FILE = PATH_TEST_DATA / 'test.osm.pbf'


# Define some shapes
coords1 = ((0., 0.), (0., 1.), (1., 1.), (1., 0.), (0., 0.)) #area=1
coords2 = ((0., 0.), (0., 0.1), (0.1, 0.1), (0.1, 0.), (0., 0.)) #area=0.01
polygon1 = sh.Polygon(coords1)
polygon2 = sh.Polygon(coords2)
point1 = sh.Point(1.0, -1.0)
point2 = sh.Point(.5, .5)
line1 = sh.LineString([[0, 0], [1, 0], [1, 1]])
line2 = sh.LineString([[0.25, 0.25], [0.75, 0.75]])

class TestSimplificationFunctions(unittest.TestCase):
    def test_remove_small_polygons(self):
        """ test function remove_small_polygons() """

        gdf_small_poly = gpd.GeoDataFrame(
            geometry = [
                point1,
                line1,
                polygon1,
                polygon2
            ])

        gdf_no_small_poly = gpd.GeoDataFrame(
            geometry = [
                point1,
                line1,
                polygon1,
            ])

        # get rid of polygon2
        gdf_removed = remove_small_polygons(gdf_small_poly, 0.5)

        assert_frame_equal(gdf_removed, gdf_no_small_poly)

    def test_remove_contained_points(self):
        """ test function remove_contained_points() """

        gdf_pnt_in_poly = gpd.GeoDataFrame(
            geometry = [polygon1, polygon2, point1, point2, line1
            ])

        gdf_check  = gpd.GeoDataFrame(
            geometry = [polygon1, polygon2, point1,line1
            ])

        # get rid of point2
        gdf_simple  = remove_contained_points(gdf_pnt_in_poly)

        assert_frame_equal(gdf_check, gdf_simple)



    def test_remove_contained_polys(self):
        """ test function remove_contained_polys() """

        gdf_poly_in_poly = gpd.GeoDataFrame(
            geometry = [polygon1, polygon2, point1, point2, line1, line2
            ])

        gdf_check  = gpd.GeoDataFrame(
            geometry = [polygon1, point1, point2, line1, line2
            ])

        # get rid of polygon2
        gdf_simple = remove_contained_polys(gdf_poly_in_poly)

        assert_frame_equal(gdf_check, gdf_simple)



    def test_remove_exact_duplicates(self):
        """ test function remove_exact_duplicates() """

        gdf_dupl = gpd.GeoDataFrame(
            geometry = [polygon1, polygon1, point1, point1, line1, line1
            ])

        # get rid of polygon1, point1, line1
        gdf_simple = remove_exact_duplicates(gdf_dupl)

        gdf_check  = gpd.GeoDataFrame(
            geometry = [polygon1, point1, line1
            ])

        assert_frame_equal(gdf_check, gdf_simple)


if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestSimplificationFunctions)
    unittest.TextTestRunner(verbosity=2).run(TESTS)


"""
This file is part of OSM-flex.
Copyright (C) 2023 OSM-flex contributors listed in AUTHORS.
OSM-flex is free software: you can redistribute it and/or modify it under the
terms of the GNU General Public License as published by the Free
Software Foundation, version 3.
OSM-flex is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.
-----
simplification functions
"""

import geopandas as gpd
import numpy as np

def remove_small_polygons(gdf, min_area):
    """Remove (multi-)polygons of area smaller than min_area
    Points and lines are untouched.

    Note: a buffer of 1e-10 is added to invalid geometries

    Parameters
    ----------
    gdf : GeoDataFrame
        geodataframe with polygons
    min_area : float
        minimal value of area

    Return
    ------
    GeoDataFrame:
        entry geodataframe without (multi-)polygons smaller than min_area
    """

    gdf_temp = gdf.copy()

    def make_valid(geometry):
        if geometry.is_valid:
            return geometry
        return geometry.buffer(1e-10)

    gdf_temp['geometry'] = gdf_temp.apply(lambda row: make_valid(row.geometry),
                                          axis=1)

    return gdf_temp[(gdf_temp['geometry'].area > min_area) |
                    (gdf_temp['geometry'].area == 0)].reset_index(drop=True)


def remove_contained_points(gdf_p_mp):
    """
    from a GeoDataFrame containing points and (multi-)polygons, remove those
    points that are contained in a multipolygons entry.
    Resets the index of the dataframe.

    Parameters
    ----------
    gdf_p_mp : gpd.GeoDataFrame
        GeoDataFrame containing entries with point and (multi-)polygon geometry
    """

    gdf_p_mp = gdf_p_mp.reset_index(drop=True)

    ind_dupl = np.unique(gpd.sjoin(gdf_p_mp[gdf_p_mp.geometry.type=='Point'],
              gdf_p_mp[(gdf_p_mp.geometry.type=='MultiPolygon')|
                       (gdf_p_mp.geometry.type=='Polygon')],
              predicate='within').index)

    return gdf_p_mp.drop(index=ind_dupl).reset_index(drop=True)


def remove_contained_polys(gdf):
    """
    from a GeoDataFrame containing (multi-)polygons (and potentially other
    geometries), remove those polygon entries that are already fully
    contained in another polygon entries. Removes smaller polygons within 
    polygons and full duplicates, but leaves contained points untouched 
    (see remove_contained_points() for this).

    Resets the index of the dataframe.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        GeoDataFrame containing entries with (multi-)polygon geometry
    """

    gdf = gdf.reset_index(drop=True)

    contained = gpd.sjoin(
        gdf[(gdf.geometry.type=='MultiPolygon')| (gdf.geometry.type=='Polygon')],
        gdf[(gdf.geometry.type=='MultiPolygon')| (gdf.geometry.type=='Polygon')],
        predicate='contains'
        )

    subset = contained[contained.index != contained.index_right]
    to_drop = set(subset.index_right) - set(subset.index)

    return gdf.drop(index=to_drop).reset_index(drop=True)


def remove_exact_duplicates(gdf):
    """
    from a GeoDataFrame containing any sort of geometries, remove those entries
    which already have an exact duplicate geometry entry.

    Resets the index of the dataframe.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        GeoDataFrame containing any types of geometry
    """

    gdf = gdf.reset_index(drop=True)

    geom_wkb = gdf["geometry"].apply(lambda geom: geom.wkb)

    return gdf.loc[geom_wkb.drop_duplicates().index].reset_index(drop=True)
